"""
Read atomic information from a PDB file. Functions and objects are
adapted from `mdtraj`.
"""

import os
import pathlib
import typing
import warnings
from copy import copy
from typing import Any, Literal, TypedDict, overload
from xml.etree import ElementTree

import jax
import mdtraj
import mmdf
import numpy as np
import pandas as pd
from jaxtyping import Float, Int
from mdtraj.core import element as elem

from ..atom_util import center_atom_positions


if hasattr(typing, "GENERATING_DOCUMENTATION"):
    AtomProperties = dict  # pyright: ignore[reportAssignmentType]
else:

    class AtomProperties(TypedDict):
        masses: Float[np.ndarray, "... n_atoms"]
        b_factors: Float[np.ndarray, "... n_atoms"]
        charges: Float[np.ndarray, "... n_atoms"]


@overload
def read_atoms_from_pdb(
    filename: str | pathlib.Path,
    *,
    loads_properties: Literal[False],
    loads_b_factors: bool = False,
    center: bool = True,
    selection_string: str = "all",
    model_index: int | None = None,
    stack_models: bool = False,
    standardizes_names: bool = True,
    topology: mdtraj.Topology | None = None,
) -> tuple[Float[np.ndarray, "... n_atoms 3"], Int[np.ndarray, " n_atoms"]]: ...


@overload
def read_atoms_from_pdb(  # type: ignore
    filename: str | pathlib.Path,
    *,
    loads_properties: Literal[True],
    loads_b_factors: bool = False,
    center: bool = True,
    selection_string: str = "all",
    model_index: int | None = None,
    stack_models: bool = False,
    standardizes_names: bool = True,
    topology: mdtraj.Topology | None = None,
) -> tuple[
    Float[np.ndarray, "... n_atoms 3"],
    Int[np.ndarray, " n_atoms"],
    AtomProperties,
]: ...


@overload
def read_atoms_from_pdb(
    filename: str | pathlib.Path,
    *,
    loads_properties: bool = False,
    loads_b_factors: bool = False,
    center: bool = True,
    selection_string: str = "all",
    model_index: int | None = None,
    stack_models: bool = False,
    standardizes_names: bool = True,
    topology: mdtraj.Topology | None = None,
) -> tuple[Float[np.ndarray, "... n_atoms 3"], Int[np.ndarray, "... n_atoms"]]: ...


def read_atoms_from_pdb(
    filename: str | pathlib.Path,
    *,
    loads_properties: bool = False,
    loads_b_factors: bool = False,
    center: bool = True,
    selection_string: str = "all",
    model_index: int | None = None,
    stack_models: bool = False,
    standardizes_names: bool = True,
    topology: mdtraj.Topology | None = None,
) -> (
    tuple[Float[np.ndarray, "... n_atoms 3"], Int[np.ndarray, "... n_atoms"]]
    | tuple[
        Float[np.ndarray, "... n_atoms 3"],
        Int[np.ndarray, " n_atoms"],
        dict[str, Any],
    ]
):
    """Load relevant atomic information for simulating cryo-EM
    images from a PDB or mmCIF file. This function wraps the function
    `mmdf_to_atoms`.

    !!! info

        The `selection_string` argument enables usage of
        [`mdtraj`](https://www.mdtraj.org/) atom selection syntax.

    **Arguments:**

    - `filename`:
        The name of the PDB/PDBx file to open.
    - `center`:
        If `True`, center the model so that its center of mass coincides
        with the origin.
    - `loads_properties`:
        If `True`, return a dictionary of the atom properties.
    - `selection_string`:
        A selection string in `mdtraj`'s format.
    - `model_index`:
        An optional index for grabbing a particular model stored in the PDB. If `None`,
        grab all models, where `atom_positions` has a leading dimension for the model
        if `stack_models = True` or concatenates all models if `stack_models = False`.
    - `stack_models`:
        If `True`, `model_index = None`, and there are multiple models in the PDB,
        assume that each model is of the same protein and return atom positions
        and properties with a stacked leading dimension.
    - `standardizes_names`:
        If `True`, non-standard atom names and residue names are standardized.
        If set to `False`, this step is skipped.
    - `topology`:
        If `None`, use the function `mmdf_to_topology` to build a topology
        on-the-fly. If `stack_models = True`, `model_index = None`, and there
        are multiple models in the PDB, use the first model index to build the
        topology.

    **Returns:**

    A tuple whose first element is a `numpy` array of coordinates containing
    atomic positions, and whose second element is an array of atomic
    numbers. To be clear,

    ```python
    atom_positons, atomic_numbers = read_atoms_from_pdb(...)
    ```

    !!! info

        If your PDB has multiple models, arrays such as the
        atom positions are loaded with a
        leading dimension for each model. To load a single
        model at index 0,

        ```python
        atom_positons, atomic_numbers = read_atoms_from_pdb(..., model_index=0)
        ```
    """
    # Load `mmdf` dataframe forward the `mmdf_to_atoms` method
    df = mmdf.read(pathlib.Path(filename))
    return mmdf_to_atoms(
        df,
        loads_properties=loads_properties,
        loads_b_factors=loads_b_factors,
        center=center,
        selection_string=selection_string,
        model_index=model_index,
        stack_models=stack_models,
        standardizes_names=standardizes_names,
        topology=topology,
    )


@overload
def mmdf_to_atoms(
    df: pd.DataFrame,
    *,
    loads_properties: Literal[False],
    loads_b_factors: bool = False,
    center: bool = True,
    selection_string: str = "all",
    model_index: int | None = None,
    stack_models: bool = False,
    standardizes_names: bool = True,
    topology: mdtraj.Topology | None = None,
) -> tuple[Float[np.ndarray, "... n_atoms 3"], Int[np.ndarray, "... n_atoms"]]: ...


@overload
def mmdf_to_atoms(  # type: ignore
    df: pd.DataFrame,
    *,
    loads_properties: Literal[True],
    loads_b_factors: bool = False,
    center: bool = True,
    selection_string: str = "all",
    model_index: int | None = None,
    stack_models: bool = False,
    standardizes_names: bool = True,
    topology: mdtraj.Topology | None = None,
) -> tuple[
    Float[np.ndarray, "... n_atoms 3"],
    Int[np.ndarray, "... n_atoms"],
    AtomProperties,
]: ...


@overload
def mmdf_to_atoms(
    df: pd.DataFrame,
    *,
    loads_properties: bool = False,
    loads_b_factors: bool = False,
    center: bool = True,
    selection_string: str = "all",
    model_index: int | None = None,
    stack_models: bool = False,
    standardizes_names: bool = True,
    topology: mdtraj.Topology | None = None,
) -> tuple[Float[np.ndarray, "... n_atoms 3"], Int[np.ndarray, "... n_atoms"]]: ...


def mmdf_to_atoms(
    df: pd.DataFrame,
    *,
    loads_properties: bool = False,
    loads_b_factors: bool = False,
    center: bool = True,
    selection_string: str = "all",
    model_index: int | None = None,
    stack_models: bool = False,
    standardizes_names: bool = True,
    topology: mdtraj.Topology | None = None,
) -> (
    tuple[Float[np.ndarray, "... n_atoms 3"], Int[np.ndarray, "... n_atoms"]]
    | tuple[
        Float[np.ndarray, "... n_atoms 3"],
        Int[np.ndarray, "... n_atoms"],
        dict[str, Any],
    ]
):
    """Load relevant atomic information for simulating cryo-EM
    images from a `pandas.DataFrame` loaded from the package
    [`mmdf`](https://github.com/teamtomo/mmdf).

    **Arguments:**

    - `df`:
        The dataframe loaded from or formatted as in
        [`mmdf`](https://github.com/teamtomo/mmdf).

    For documentation of other arguments and return value,
    see the function `read_atoms_from_pdb`.
    ```
    """
    # Load atom info from `mmdf` dataframe
    atom_info = _load_atom_info(df, model_index=model_index, stack_models=stack_models)
    if selection_string != "all":
        if topology is None:
            if model_index is None and stack_models:
                topology = mmdf_to_topology(df, standardizes_names, model_index=0)
            else:
                topology = mmdf_to_topology(
                    df, standardizes_names, model_index=model_index
                )
        # Filter atoms and grab atomic positions and numbers
        selected_indices = topology.select(selection_string)
        atom_positions = atom_info["positions"][:, selected_indices]
        atomic_numbers = atom_info["numbers"][selected_indices]
        atom_properties = jax.tree.map(
            lambda arr: arr[:, selected_indices], atom_info["properties"]
        )
    else:
        atom_positions = atom_info["positions"]
        atomic_numbers = atom_info["numbers"]
        atom_properties = atom_info["properties"]
    # Center by mass
    if center:
        atom_masses = atom_properties["masses"]
        atom_positions = center_atom_positions(atom_positions, atom_masses, atom_axis=1)
    # Return, without leading dimensions if there is only one structure
    atom_positions = atom_positions[0] if atom_positions.shape[0] == 1 else atom_positions
    if loads_properties or loads_b_factors:
        # Optionally return atom properties
        atom_properties = jax.tree.map(
            lambda arr: (arr[0] if arr.shape[0] == 1 else arr), atom_properties
        )
        if loads_b_factors:
            warnings.warn(
                "`loads_b_factor` option is deprecated and will be removed in "
                "cryoJAX 0.6.0. Use `loads_properties` instead.",
                category=FutureWarning,
                stacklevel=2,
            )
            return atom_positions, atomic_numbers, atom_properties["b_factors"]
        else:
            return atom_positions, atomic_numbers, atom_properties
    else:
        return atom_positions, atomic_numbers


def mmdf_to_topology(
    df: pd.DataFrame,
    standardizes_names: bool = True,
    model_index: int | None = None,
) -> mdtraj.Topology:
    """Generate an `mdtraj.Topology` using an array of atom
    positions and a `pandas.DataFrame` loaded from the package
    [`mmdf`](https://github.com/teamtomo/mmdf).

    **Arguments:**

    - `df`:
        The dataframe loaded from or formatted as in
        [`mmdf`](https://github.com/teamtomo/mmdf).
    - `standardizes_names`:
        If `True`, non-standard atom names and residue names are
        standardized.
    - `model_index`:
        The model index from which to build the topology. Possible
        indices are captured in `df["model"]`. If `None`, use all
        models to build a single topology.

    **Returns:**

    An `mdtraj.Topology` object.
    """
    topology = mdtraj.Topology()
    if standardizes_names:
        residue_name_replacements, atom_name_replacements = (
            _load_name_replacement_tables()
        )
    else:
        residue_name_replacements, atom_name_replacements = {}, {}
    df_at_model = df if model_index is None else df[df["model"] == model_index]
    for atom_index in range(len(df_at_model)):
        df_at_index = df_at_model.iloc[atom_index]
        chain_id = df_at_index["chain"]
        residue_name = df_at_index["residue"]
        residue_id = df_at_index["residue_id"]
        c = topology.add_chain(chain_id)
        if residue_name in residue_name_replacements and standardizes_names:
            residue_name = residue_name_replacements[residue_name]
        # TODO: is it necessary to have `segment_id`, as is parsed in `mdtraj`?
        r = topology.add_residue(residue_name, c, residue_id, segment_id="")
        if residue_name in atom_name_replacements and standardizes_names:
            atom_replacements = atom_name_replacements[residue_name]
        else:
            atom_replacements = {}
        atom_name = df_at_index["atom"]
        if atom_name in atom_replacements:
            atom_name = atom_replacements[atom_name]
        atom_name = atom_name.strip()
        element = elem.Element.getByAtomicNumber(df_at_index["atomic_number"])
        charges = df_at_index["charge"]
        # TODO: ok to remove serial number?
        _ = topology.add_atom(
            atom_name,
            element,
            r,
            serial=atom_index,  # atom.serial_number,
            formal_charge=charges,
        )
    # Generate bonds
    atom_positions = df_at_model[["x", "y", "z"]].to_numpy()
    topology.create_standard_bonds()
    topology.create_disulfide_bonds(atom_positions.tolist())

    return topology


def read_topology_from_pdb(
    filename: str | pathlib.Path,
    model_index: int | None = None,
    standardizes_names: bool = True,
) -> mdtraj.Topology:
    """Generate an `mdtraj.Topology` from a PDB or mmCIF file.
    This function wraps the function `mmdf_to_topology`.

    !!! info
        Since we use `mmdf` to parse the PDB/PDBx file, the
        atom ordering in some of our functions, e.g., `read_atoms_from_pdb`
        may differ from that of `mdtraj.load`. We recommend using this function
        if you need a topology that is consistent with that of `read_atoms_from_pdb`.

    **Arguments:**

    - `df`:
        The dataframe loaded from or formatted as in
        [`mmdf`](https://github.com/teamtomo/mmdf).
    - `standardizes_names`:
        If `True`, non-standard atom names and residue names are
        standardized.
    - `model_index`:
        The model index from which to build the topology. Possible
        indices are captured in `df["model"]`.

    **Returns:**

    An `mdtraj.Topology` object.
    """
    df = mmdf.read(pathlib.Path(filename))
    return mmdf_to_topology(df, standardizes_names, model_index)


class _AtomicModelInfo(TypedDict):
    positions: Float[np.ndarray, "M N 3"]
    numbers: Int[np.ndarray, "M N 3"]
    properties: AtomProperties


def _load_atom_info(df: pd.DataFrame, model_index: int | None, stack_models: bool):
    if df.size == 0:
        raise ValueError(
            "When loading an atomic model using `mmdf`, found that "
            "the dataframe was empty."
        )
    # Load atom info
    if model_index is not None:
        df = df[df["model"] == model_index]
        if df.size == 0:
            raise ValueError(
                f"Found no atoms matching `model_index = {model_index}`. "
                "Model numbers available for indexing are "
                f"{df['model'].unique().tolist()}. "
            )
    if model_index is None and stack_models:
        model_numbers = df["model"].unique().tolist()
    else:
        model_numbers = [None]
    atom_positions, atomic_numbers = [], []
    atom_masses, b_factors, charges = [], [], []
    for index in model_numbers:
        df_at_model = df if index is None else df[df["model"] == index]
        atom_positions.append(df_at_model[["x", "y", "z"]].to_numpy())
        atomic_numbers.append(df_at_model["atomic_number"].to_numpy())
        atom_masses.append(df_at_model["atomic_weight"].to_numpy())
        b_factors.append(df_at_model["b_isotropic"].to_numpy())
        charges.append(df_at_model["charge"].to_numpy())

    if len(atomic_numbers) > 1:
        if not all(np.array_equal(arr, atomic_numbers[0]) for arr in atomic_numbers):
            raise ValueError(
                "Tried to load multiple models with `stack_models = True`, "
                "but found that atomic numbers across different models "
                "were different. Only use `stack_models = True` if "
                "each model represents the same protein."
            )

    # Gather atom info and return
    properties = AtomProperties(
        charges=np.asarray(charges, dtype=int),
        b_factors=np.asarray(b_factors, dtype=float),
        masses=np.asarray(atom_masses, dtype=float),
    )
    atom_info = _AtomicModelInfo(
        positions=np.asarray(atom_positions, dtype=float),
        numbers=np.asarray(atomic_numbers[0], dtype=int),
        properties=properties,
    )

    return atom_info


def _load_name_replacement_tables():
    """Load the list of atom and residue name replacements.
    Closely follows `mdtraj.formats.pdb.PDBTrajectoryFile._loadNameReplacementTables`.
    """
    tree = ElementTree.parse(
        os.path.join(os.path.dirname(__file__), "pdb_names.xml"),
    )
    # Residue and atom replacements
    residue_name_replacements = {}
    atom_name_replacements = {}
    # ... containers for residues
    all_residues, protein_residues, nucleic_acid_residues = {}, {}, {}
    for residue in tree.getroot().findall("Residue"):
        name = residue.attrib["name"]
        if name == "All":
            _parse_residue(residue, all_residues)
        elif name == "Protein":
            _parse_residue(residue, protein_residues)
        elif name == "Nucleic":
            _parse_residue(residue, nucleic_acid_residues)
    for atom in all_residues:
        protein_residues[atom] = all_residues[atom]
        nucleic_acid_residues[atom] = all_residues[atom]
    for residue in tree.getroot().findall("Residue"):
        name = residue.attrib["name"]
        for id in residue.attrib:
            if id == "name" or id.startswith("alt"):
                residue_name_replacements[residue.attrib[id]] = name
        if "type" not in residue.attrib:
            atoms = copy(all_residues)
        elif residue.attrib["type"] == "Protein":
            atoms = copy(protein_residues)
        elif residue.attrib["type"] == "Nucleic":
            atoms = copy(nucleic_acid_residues)
        else:
            atoms = copy(all_residues)
        _parse_residue(residue, atoms)
        atom_name_replacements[name] = atoms
    return residue_name_replacements, atom_name_replacements


def _parse_residue(residue, map):
    """Closely follows `mdtraj.formats.pdb.PDBTrajectoryFile._parseResidueAtoms`."""
    for atom in residue.findall("Atom"):
        name = atom.attrib["name"]
        for id in atom.attrib:
            map[atom.attrib[id]] = name



import typing
from typing import TypeAlias

import numpy as np
from jaxtyping import Array, Bool, Complex, Float, Inexact, Int


if hasattr(typing, "GENERATING_DOCUMENTATION"):

    class NDArrayLike:  # pyright: ignore[reportRedeclaration]
        """A type hint for a JAX or numpy array"""

        pass

    class BoolLike:  # pyright: ignore[reportRedeclaration]
        """A type hint for a python `bool` or a
        JAX/numpy boolean scalar.
        """

        pass

    class ComplexLike:  # pyright: ignore[reportRedeclaration]
        """A type hint for a python `complex` or a
        JAX/numpy complex scalar.
        """

        pass

    class FloatLike:  # pyright: ignore[reportRedeclaration]
        """A type hint for a python `float` or a
        JAX/numpy float scalar.
        """

        pass

    class InexactLike:  # pyright: ignore[reportRedeclaration]
        """A type hint for a python `float` / `complex`, or a
        JAX/numpy float / complex scalar.
        """

        pass

    class IntLike:  # pyright: ignore[reportRedeclaration]
        """A type hint for a python `int`, or a
        JAX/numpy integer scalar.
        """

        pass

else:
    NDArrayLike: TypeAlias = Array | np.ndarray
    BoolLike: TypeAlias = bool | Bool[Array | np.ndarray, ""]
    ComplexLike: TypeAlias = complex | Complex[Array | np.ndarray, ""]
    FloatLike: TypeAlias = float | Float[Array | np.ndarray, ""]
    InexactLike: TypeAlias = complex | float | Inexact[Array | np.ndarray, ""]
    IntLike: TypeAlias = int | Int[Array | np.ndarray, ""]



from collections.abc import Callable
from typing import Any

import equinox as eqx
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import Array, Int, PyTree

from .custom_types import PyTreeGrid, PyTreeGridIndex, PyTreeGridPoint


def tree_grid_shape(
    tree_grid: PyTreeGrid,
    *,
    is_leaf: Callable[[Any], bool] | None = None,
) -> tuple[int, ...]:
    """Get the shape of a pytree grid.

    **Arguments:**

    - `tree_grid`: A sparse grid cartesian grid, represented as a pytree.
                   See [`run_grid_search`][] for more information.
    - `is_leaf`: As [`jax.tree_util.tree_flatten`](https://jax.readthedocs.io/en/latest/_autosummary/jax.tree_util.tree_flatten.html).

    **Returns:**

    The shape of `tree_grid`.

    !!! Example

        ```python
        # A simple "pytree grid"
        simple_tree_grid = (jnp.zeros(10), jnp.zeros(10), jnp.zeros(10))
        # Its shape is just the shape of the cartesian product of its leaves
        assert tree_grid_shape(simple_tree_grid) == (10, 10, 10)
        ```

    !!! Example

        ```python
        # Library code
        import equinox as eqx
        import jax

        class SomeModule(eqx.Module):

            a: jax.Array

        # End-user script
        # ... create a more complicated grid
        complicated_tree_grid = (SomeModule(jnp.zeros(10)), jnp.zeros(10), (jnp.zeros(10), None))
        # Its shape is still just the shape of the cartesian product of its leaves
        assert tree_grid_shape(complicated_tree_grid) == (10, 10, 10)
        ```
    """  # noqa: E501
    n_leaves = len(jtu.tree_leaves(tree_grid, is_leaf=is_leaf))
    if n_leaves == 0:
        raise ValueError(
            "The pytree passed to `tree_grid_shape` should have at least "
            f"one leaf. The pytree was equal to {tree_grid}, which has "
            "no leaves."
        )
    else:
        _leading_dim_resolver = jtu.tree_map(
            _LeafLeadingDimension, tree_grid, is_leaf=is_leaf
        )
        _reduce_fn = lambda x, y: (
            x.get() + y.get() if isinstance(x, _LeafLeadingDimension) else x + y.get()
        )
        shape = jtu.tree_reduce(
            _reduce_fn,
            _leading_dim_resolver,
            is_leaf=lambda x: isinstance(x, _LeafLeadingDimension),
        )
        return shape if n_leaves > 1 else shape.get()


def tree_grid_unravel_index(
    raveled_index: int | Int[Array, ""] | Int[Array, " _"],
    tree_grid: PyTreeGrid,
    *,
    is_leaf: Callable[[Any], bool] | None = None,
) -> PyTreeGridIndex:
    """Get a "grid index" for a pytree grid.

    Roughly, this can be thought of as `jax.numpy.unravel_index`, but with a
    pytree grid. See [`tree_grid_take`][] for an example of how to use this
    function to sample a grid point.

    **Arguments:**

    - `raveled_index`: A flattened index for `tree_grid`. Simply pass an integer
                       valued index, as one would with a flattened array. Passing
                       a 1D array of indices is also supported.
    - `tree_grid`: A sparse grid cartesian grid, represented as a pytree.
                   See [`run_grid_search`][] for more information.
    - `is_leaf`: As [`jax.tree_util.tree_flatten`](https://jax.readthedocs.io/en/latest/_autosummary/jax.tree_util.tree_flatten.html).

    **Returns:**

    The grid index. This is a pytree of the same structure as `tree_grid`, with the
    result of `jax.numpy.unravel_index(raveled_index, shape)` inserted into the
    appropriate leaf. Here, `shape` is given by the output of [`tree_grid_shape`][].
    """
    raveled_index = jnp.asarray(raveled_index)
    shape = tree_grid_shape(tree_grid, is_leaf=is_leaf)
    # raveled_index = eqx.error_if(
    #     raveled_index,
    #     jnp.logical_or(raveled_index < 0, raveled_index >= math.prod(shape)),
    #     "The flattened grid index must be greater than 0 and less than the "
    #     f"grid size. Got index {raveled_index}, but the grid has shape {shape}, "
    #     f"so its maximum index is {math.prod(shape) - 1}.",
    # )
    unraveled_index = jnp.unravel_index(raveled_index, shape)
    tree_grid_def = jtu.tree_structure(tree_grid, is_leaf=is_leaf)
    tree_grid_index = jtu.tree_unflatten(tree_grid_def, unraveled_index)

    return tree_grid_index


def tree_grid_take(
    tree_grid: PyTreeGrid,
    tree_grid_index: PyTreeGridIndex,
) -> PyTreeGridPoint:
    """Get a grid point of the pytree grid, given a
    pytree grid index. See [`tree_grid_unravel_index`][] to see
    how to return a pytree grid index.

    Roughly, this can be thought of as `jax.numpy.take`, but with a
    pytree grid.

    **Arguments:**

    - `tree_grid`: A sparse cartesian grid, represented as a pytree.
                   See [`run_grid_search`][] for more information.
    - `tree_grid_index`: An index for `tree_grid`, also represented as a pytree.
                   See [`tree_grid_unravel_index`][] for more information.

    **Returns:**

    A grid point of a pytree grid. This is a pytree of the same
    structure as `tree_grid` (or a prefix of it), where each leaf
    is indexed by the leaf at `tree_grid_index`.

    !!! Example

        ```python
        # A simple "pytree grid"
        simple_tree_grid = (jnp.zeros(10), jnp.zeros(10), jnp.zeros(10))
        # Its shape is just the shape of the cartesian product of its leaves
        raveled_index = 7
        tree_grid_index = tree_grid_unravel_index(raveled_index, simple_tree_grid)
        tree_grid_point = tree_grid_take(simple_tree_grid, tree_grid_index)
        assert tree_grid_point == (jnp.asarray(0.), jnp.asarray(0.), jnp.asarray(0.))
        ```
    """
    tree_grid_point = _tree_take(tree_grid, tree_grid_index, axis=0)
    return tree_grid_point


def _tree_take(
    pytree_of_arrays: PyTree[Array],
    pytree_of_indices: PyTree[Int[Array, "..."]],
    axis: int | None = None,
    mode: str | None = None,
    fill_value: Array | None = None,
) -> PyTree[Array]:
    return jtu.tree_map(
        lambda i, l: _leaf_take(i, l, axis=axis, mode=mode, fill_value=fill_value),
        pytree_of_indices,
        pytree_of_arrays,
    )


def _leaf_take(index, leaf, **kwargs):
    _take_fn = lambda array: jnp.take(jnp.atleast_1d(array), index, **kwargs)
    if eqx.is_array(leaf):
        return _take_fn(leaf)
    else:
        return jtu.tree_map(_take_fn, leaf)


def _get_leading_dim(array):
    return (array.shape[0],)


class _LeafLeadingDimension(eqx.Module):
    _leaf: Any

    def get(self):
        if eqx.is_array(self._leaf):
            return _get_leading_dim(self._leaf)
        else:
            leaves = jtu.tree_leaves(self._leaf)
            if len(leaves) > 0:
                _leading_dim = _get_leading_dim(leaves[0])
                if not all([_get_leading_dim(leaf) == _leading_dim for leaf in leaves]):
                    raise ValueError(
                        "Arrays stored in PyTree leaves should share the same "
                        "leading dimension. Found that this is not true for "
                        f"leaf {self._leaf}."
                    )
                return _leading_dim
            else:
                raise ValueError(f"No arrays found at leaf {self._leaf}")



import jax.numpy as jnp
from jaxtyping import Array, Float

from ...jax_util import FloatLike
from ...ndimage import cartesian_to_polar


# Not currently public API
def compute_phase_shifts_with_spherical_aberration(
    frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"],
    defocus_in_angstroms: Float[Array, ""],
    astigmatism_in_angstroms: Float[Array, ""],
    astigmatism_angle: Float[Array, ""],
    wavelength_in_angstroms: Float[Array, ""],
    spherical_aberration_in_angstroms: Float[Array, ""],
) -> Float[Array, "y_dim x_dim"]:
    k_sqr, azimuth = cartesian_to_polar(frequency_grid_in_angstroms, square=True)
    defocus = defocus_in_angstroms + 0.5 * astigmatism_in_angstroms * jnp.cos(
        2.0 * (azimuth - astigmatism_angle)
    )
    defocus_phase_shifts = -0.5 * defocus * wavelength_in_angstroms * k_sqr
    aberration_phase_shifts = (
        0.25
        * spherical_aberration_in_angstroms
        * (wavelength_in_angstroms**3)
        * (k_sqr**2)
    )
    phase_shifts = (2 * jnp.pi) * (defocus_phase_shifts + aberration_phase_shifts)

    return phase_shifts


# Not currently public API
def compute_phase_shift_from_amplitude_contrast_ratio(
    amplitude_contrast_ratio: FloatLike,
) -> Float[Array, ""]:
    amplitude_contrast_ratio = jnp.asarray(amplitude_contrast_ratio, dtype=float)
    return jnp.arctan(
        amplitude_contrast_ratio / jnp.sqrt(1.0 - amplitude_contrast_ratio**2)
    )



import warnings as _warnings
from typing import Any as _Any

from ._dataset import AbstractDataset as AbstractDataset
from ._particle_data import (
    AbstractParticleDataset as AbstractParticleDataset,
    AbstractParticleParameterFile as AbstractParticleParameterFile,
    AbstractParticleStarFile as AbstractParticleStarFile,
    RelionParticleDataset as RelionParticleDataset,
    RelionParticleParameterFile as RelionParticleParameterFile,
    simulate_particle_stack as simulate_particle_stack,
)


_warnings.warn(
    "Submodule `cryojax.dataset` is deprecated and "
    "has been moved to the library `cryospax` "
    "(https://github.com/michael-0brien/cryospax). "
    "`cryojax.dataset` will be removed in cryoJAX 0.6.0.",
    category=FutureWarning,
    stacklevel=2,
)


def __getattr__(name: str) -> _Any:
    # Future deprecations
    if name == "RelionParticleStackDataset":
        _warnings.warn(
            "The 'RelionParticleStackDataset' alias is deprecated and will be removed in "
            "`cryospax`. Use 'RelionParticleDataset' instead.",
            category=FutureWarning,
            stacklevel=2,
        )
        return RelionParticleDataset

    if name == "ParticleParameterInfo":
        _warnings.warn(
            "The 'ParticleParameterInfo' TypedDict is deprecated and will be removed in "
            "`cryospax`.",
            category=FutureWarning,
            stacklevel=2,
        )
        from ._particle_data.relion import (
            _ParticleParameterInfo as ParticleParameterInfo,
        )

        return ParticleParameterInfo

    if name == "ParticleStackInfo":
        _warnings.warn(
            "The 'ParticleStackInfo' TypedDict is deprecated and will be removed in "
            "`cryospax`.",
            category=FutureWarning,
            stacklevel=2,
        )
        from ._particle_data.relion import (
            _ParticleStackInfo as ParticleStackInfo,
        )

        return ParticleStackInfo



from abc import abstractmethod
from typing import Generic, TypeVar
from typing_extensions import override

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import Array, Complex

from ...ndimage import fftn, ifftn, map_coordinates, resize_with_crop_or_pad
from .._image_config import AbstractImageConfig
from .._volume import AbstractVolumeRepresentation, RealVoxelGridVolume


VolRep = TypeVar("VolRep", bound="AbstractVolumeRepresentation")


class AbstractMultisliceIntegrator(eqx.Module, Generic[VolRep], strict=True):
    """Base class for a multislice integration scheme."""

    @abstractmethod
    def integrate(
        self, volume_representation: VolRep, image_config: AbstractImageConfig
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]:
        raise NotImplementedError


class FFTMultisliceIntegrator(
    AbstractMultisliceIntegrator[RealVoxelGridVolume], strict=True
):
    """Multislice integrator that steps using successive FFT-based convolutions."""

    slice_thickness_in_voxels: int

    def __init__(
        self,
        slice_thickness_in_voxels: int = 1,
    ):
        """**Arguments:**

        - `slice_thickness_in_voxels`:
            The number of slices to step through per iteration of the
            rasterized voxel grid.
        - `options_for_interpolation`:
            See `cryojax.image.map_coordinates` for documentation.
        """
        if slice_thickness_in_voxels < 1:
            raise AttributeError(
                "FFTMultisliceIntegrator.slice_thickness_in_voxels must be an "
                "integer greater than or equal to 1."
            )
        self.slice_thickness_in_voxels = slice_thickness_in_voxels

    @override
    def integrate(
        self,
        volume_representation: RealVoxelGridVolume,
        image_config: AbstractImageConfig,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]:
        """Compute the exit wave from an atomic potential using the multislice
        method.

        **Arguments:**

        - `volume_representation`:
            The volume to integrate to the exit plane. This is
            a real-valued voxel grid, which must be in physical units
            of a scattering potential. See rendering method
            `PengAtomPotential.to_real_voxel_grid` for an example.
        - `image_config`:
            The configuration of the imaging instrument.

        **Returns:**

        The wavefunction in the exit plane of the specimen.
        """  # noqa: E501
        # Interpolate volume to new pose at given coordinate system
        z_dim, y_dim, x_dim = volume_representation.shape
        voxel_size = image_config.pixel_size
        real_voxel_grid = _sample_voxel_grid(volume_representation)
        # Initialize multislice geometry
        n_slices = z_dim // self.slice_thickness_in_voxels
        slice_thickness = voxel_size * self.slice_thickness_in_voxels
        # Locally average the potential to be at the given slice thickness.
        # Thow away some slices equal to the remainder
        # `dim % self.slice_thickness_in_voxels`
        if self.slice_thickness_in_voxels > 1:
            potential_per_slice = jnp.sum(
                real_voxel_grid[
                    : z_dim - z_dim % self.slice_thickness_in_voxels, ...
                ].reshape((self.slice_thickness_in_voxels, n_slices, y_dim, x_dim)),
                axis=0,
            )
            # ... take care of remainder
            if z_dim % self.slice_thickness_in_voxels != 0:
                potential_per_slice = jnp.concatenate(
                    (
                        potential_per_slice,
                        real_voxel_grid[
                            z_dim - z_dim % self.slice_thickness_in_voxels :, ...
                        ],
                    )
                )
        else:
            potential_per_slice = real_voxel_grid
        # Compute the integrated potential in a given slice interval, multiplying by
        # the slice thickness (TODO: interpolate for different slice thicknesses?)
        const = image_config.interaction_constant * voxel_size
        compute_object_fn = lambda pot: const * pot
        object_per_slice = jax.vmap(compute_object_fn)(potential_per_slice)
        # Compute the transmission function
        transmission = jnp.exp(1.0j * object_per_slice)
        # Compute the fresnel propagator (TODO: check numerical factors)
        q = image_config.get_frequency_grid(padding=True, physical=True, full=True)
        q_sqr = jnp.sum(q**2, axis=-1)
        fresnel_propagator = jnp.exp(
            1.0j * jnp.pi * image_config.wavelength_in_angstroms * q_sqr * slice_thickness
        )
        # Prepare for iteration. First, initialize plane wave
        plane_wave = jnp.ones((y_dim, x_dim), dtype=complex)
        # ... stepping function
        make_step = lambda n, last_exit_wave: ifftn(
            fftn(transmission[n, :, :] * last_exit_wave) * fresnel_propagator
        )
        # Compute exit wave
        exit_wave = jax.lax.fori_loop(0, n_slices, make_step, plane_wave)
        # Resize the image to match the AbstractImageConfig.padded_shape
        if image_config.padded_shape != exit_wave.shape:
            exit_wave = resize_with_crop_or_pad(
                exit_wave, image_config.padded_shape, constant_values=1.0 + 0.0j
            )

        return exit_wave


def _sample_voxel_grid(volume: RealVoxelGridVolume):
    # Convert to logical coordinates
    z_dim, y_dim, x_dim = volume.real_voxel_grid.shape
    logical_coordinate_grid = (
        volume.coordinate_grid_in_pixels
        + jnp.asarray((x_dim // 2, y_dim // 2, z_dim // 2))[None, None, None, :]
    )
    # Convert arguments to map_coordinates convention and compute
    x, y, z = jnp.transpose(logical_coordinate_grid, axes=[3, 0, 1, 2])
    return map_coordinates(volume.real_voxel_grid, (z, y, x), order=1, mode="fill")



import cryojax.simulator as cxs
import jax.random as jr
import numpy as np
import pytest
from cryojax.io import read_array_from_mrc


@pytest.fixture
def volume_and_pixel_size(sample_mrc_path):
    real_voxel_grid, voxel_size = read_array_from_mrc(
        sample_mrc_path, loads_grid_spacing=True
    )
    return (
        cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxel_grid, pad_scale=1.3),
        voxel_size,
    )


@pytest.fixture
def volume(volume_and_pixel_size):
    return volume_and_pixel_size[0]


@pytest.fixture
def basic_config(volume_and_pixel_size):
    volume, pixel_size = volume_and_pixel_size
    return cxs.BasicImageConfig(
        shape=volume.shape[0:2],
        pixel_size=pixel_size,
        voltage_in_kilovolts=300.0,
    )


@pytest.fixture
def dose_config(volume_and_pixel_size):
    volume, pixel_size = volume_and_pixel_size
    return cxs.DoseImageConfig(
        shape=volume.shape[0:2],
        pixel_size=pixel_size,
        voltage_in_kilovolts=300.0,
        electron_dose=100.0,
    )


@pytest.fixture
def image_model(volume, basic_config):
    image_model = cxs.make_image_model(
        volume,
        basic_config,
        pose=cxs.EulerAnglePose(),
        transfer_theory=cxs.ContrastTransferTheory(cxs.AstigmaticCTF()),
    )
    return image_model


@pytest.fixture
def detector_image_model(volume, dose_config):
    detector = cxs.PoissonDetector(cxs.NullDQE())
    scattering_theory = cxs.WeakPhaseScatteringTheory(
        volume_integrator=cxs.FourierSliceExtraction(),
        transfer_theory=cxs.ContrastTransferTheory(cxs.AstigmaticCTF()),
    )
    image_model = cxs.ElectronCountsImageModel(
        volume,
        image_config=dose_config,
        pose=cxs.EulerAnglePose(),
        scattering_theory=scattering_theory,
        detector=detector,
        normalizes_signal=True,
    )
    return image_model


@pytest.fixture
def solvent_image_model(volume, basic_config):
    import cryojax.experimental as cxe

    scattering_theory = cxs.WeakPhaseScatteringTheory(
        volume_integrator=cxs.FourierSliceExtraction(),
        transfer_theory=cxs.ContrastTransferTheory(cxs.AstigmaticCTF()),
        solvent=cxe.GRFSolvent2D(100.0),
    )
    image_model = cxs.ContrastImageModel(
        volume,
        image_config=basic_config,
        pose=cxs.EulerAnglePose(),
        scattering_theory=scattering_theory,
        normalizes_signal=True,
    )
    return image_model


@pytest.mark.parametrize(
    "cls, model",
    [
        (cxs.GaussianWhiteNoiseModel, "image_model"),
        (cxs.GaussianColoredNoiseModel, "image_model"),
    ],
)
def test_simulate_equals_compute_signal(cls, model, request):
    model = request.getfixturevalue(model)
    noise_model = cls(model)
    np.testing.assert_allclose(model.simulate(), noise_model.compute_signal())


@pytest.mark.parametrize("variance, dim", ((1.0, 150), (2.0, 150), (4.0, 150)))
def test_variance_correct(variance, dim):
    rng_seed = 123
    voxel_volume = cxs.FourierVoxelGridVolume.from_real_voxel_grid(
        np.zeros((dim, dim, dim))
    )
    image_config = cxs.BasicImageConfig(
        (dim, dim), pixel_size=1.0, voltage_in_kilovolts=300.0
    )
    image_model = cxs.make_image_model(
        voxel_volume, image_config, pose=cxs.EulerAnglePose()
    )
    rng_key = jr.key(rng_seed)
    noise_model = cxs.GaussianWhiteNoiseModel(image_model, variance=variance)
    np.testing.assert_allclose(
        noise_model.compute_noise(rng_key).std(), np.sqrt(variance), rtol=1e-2
    )


def test_detector_incompatible(detector_image_model):
    with pytest.raises(ValueError):
        _ = cxs.GaussianWhiteNoiseModel(detector_image_model)


def test_solvent2d_incompatible(solvent_image_model):
    with pytest.raises(ValueError):
        _ = cxs.GaussianWhiteNoiseModel(solvent_image_model)



"""
Routines for working with MRC files.
"""

import pathlib
from typing import Literal, cast, overload

import mrcfile
import numpy as np
from jaxtyping import Array, Float


@overload
def read_array_from_mrc(
    filename: str | pathlib.Path,
    *,
    loads_grid_spacing: Literal[False],
    mmap: bool = False,
    permissive: bool = False,
) -> Float[np.ndarray, "..."]: ...


@overload
def read_array_from_mrc(  # type: ignore
    filename: str | pathlib.Path,
    *,
    loads_grid_spacing: Literal[True],
    mmap: bool = False,
    permissive: bool = False,
) -> tuple[Float[np.ndarray, "..."], float]: ...


@overload
def read_array_from_mrc(
    filename: str | pathlib.Path,
    *,
    loads_grid_spacing: bool = False,
    mmap: bool = False,
    permissive: bool = False,
) -> Float[np.ndarray, "..."]: ...


def read_array_from_mrc(
    filename: str | pathlib.Path,
    *,
    loads_grid_spacing: bool = False,
    mmap: bool = False,
    permissive: bool = False,
) -> Float[np.ndarray, "..."] | tuple[Float[np.ndarray, "..."], float]:
    """Read an MRC data into a a numpy array.

    **Arguments:**

    - `filename` : Path to data.
    - `mmap`: Whether or not to open the data as a `numpy.memmap` array.
    - `loads_grid_spacing`: If `True`, load the pixel or voxel size of the array.

    **Returns:**

    - 'array' : The array stored in the Mrcfile.
    """
    # Validate filename as MRC path
    _ = _validate_filename_and_return_suffix(filename)
    # Read MRC
    open = mrcfile.mmap if mmap else mrcfile.open
    with open(filename, mode="r", permissive=permissive) as mrc:
        array = cast(np.ndarray, mrc.data)
        if mrc.is_single_image():
            grid_spacing_per_dimension = np.asarray(
                [mrc.voxel_size.x, mrc.voxel_size.y], dtype=float
            )
        elif mrc.is_image_stack():
            grid_spacing_per_dimension = np.asarray(
                [mrc.voxel_size.x, mrc.voxel_size.y], dtype=float
            )
        elif mrc.is_volume():
            grid_spacing_per_dimension = np.asarray(
                [mrc.voxel_size.x, mrc.voxel_size.y, mrc.voxel_size.z],
                dtype=float,
            )
        elif mrc.is_volume_stack():
            grid_spacing_per_dimension = np.asarray(
                [mrc.voxel_size.x, mrc.voxel_size.y, mrc.voxel_size.z],
                dtype=float,
            )
        else:
            raise ValueError(
                "Mrcfile could not be identified as an image, image stack, volume, or "
                f"volume stack. Run mrcfile.validate(...) to make sure {filename} is a "
                "valid MRC file."
            )

        if not mmap:
            array = array.astype(np.float64)

        if loads_grid_spacing:
            # Only allow the same spacing in each direction
            assert all(
                grid_spacing_per_dimension != np.zeros(grid_spacing_per_dimension.shape)
            ), "Mrcfile.voxel_size must be set if reading the grid spacing. Found that "
            "Mrcfile.voxel_size = (0.0, 0.0, 0.0)"
            assert all(grid_spacing_per_dimension == grid_spacing_per_dimension[0]), (
                "Mrcfile.voxel_size must be same in all dimensions."
            )
            grid_spacing = grid_spacing_per_dimension[0]

            return cast(np.ndarray, array), float(grid_spacing)
        else:
            # ... otherwise, just return
            return cast(np.ndarray, array)


def write_image_to_mrc(
    image: Float[Array, "y_dim x_dim"] | Float[np.ndarray, "y_dim x_dim"],
    pixel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
    filename: str | pathlib.Path,
    overwrite: bool = False,
    compression: str | None = None,
):
    """Write an image stack to an MRC file.

    **Arguments:**

    - `image`: The image stack, where the leading dimension indexes the image.
    - `pixel_size`: The pixel size of the images in `image_stack`.
    - `filename`: The output filename.
    - `overwrite`: If `True`, overwrite an existing file.
    - `compression`: See `mrcfile.new` for documentation.
    """
    # Validate filename as MRC path and get suffix
    suffix = _validate_filename_and_return_suffix(filename)
    if suffix != ".mrc":
        raise OSError(
            f"The suffix for an image in MRC format must be .mrc. Instead, got {suffix}."
        )
    if image.ndim != 2:
        raise ValueError("image.ndim must be equal to 2.")
    # Convert image stack and pixel size to numpy arrays.
    image_as_numpy = np.asarray(image)
    pixel_size_as_numpy = np.asarray(pixel_size)
    # Create new file and write
    with mrcfile.new(filename, compression=compression, overwrite=overwrite) as mrc:
        mrc.set_data(image_as_numpy)
        mrc.voxel_size = pixel_size_as_numpy


def write_image_stack_to_mrc(
    image_stack: Float[Array, "M y_dim x_dim"] | Float[np.ndarray, "M y_dim x_dim"],
    pixel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
    filename: str | pathlib.Path,
    overwrite: bool = False,
    compression: str | None = None,
):
    """Write an image stack to an MRC file.

    **Arguments:**

    - `image_stack`: The image stack, where the leading dimension indexes the image.
    - `pixel_size`: The pixel size of the images in `image_stack`.
    - `filename`: The output filename.
    - `overwrite`: If `True`, overwrite an existing file.
    - `compression`: See `mrcfile.new` for documentation.
    """
    # Validate filename as MRC path and get suffix
    suffix = _validate_filename_and_return_suffix(filename)
    if suffix != ".mrcs":
        raise OSError(
            "The suffix for an image stack in MRC format must be .mrcs. "
            f"Instead, got {suffix}."
        )
    if image_stack.ndim != 3:
        raise ValueError("image_stack.ndim must be equal to 3.")
    # Convert image stack and pixel size to numpy arrays.
    image_stack_as_numpy = np.asarray(image_stack)
    pixel_size_as_numpy = np.asarray(pixel_size)
    # Create new file and write
    with mrcfile.new(filename, compression=compression, overwrite=overwrite) as mrc:
        mrc.set_data(image_stack_as_numpy)
        mrc.set_image_stack()
        mrc.voxel_size = pixel_size_as_numpy


def write_volume_to_mrc(
    voxel_grid: (
        Float[Array, "z_dim y_dim x_dim"] | Float[np.ndarray, "z_dim y_dim x_dim"]
    ),
    voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
    filename: str | pathlib.Path,
    overwrite: bool = False,
    compression: str | None = None,
):
    """Write a voxel grid to an MRC file.

    **Arguments:**

    - `voxel_grid`: The voxel grid as a JAX array.
    - `voxel_size`: The voxel size of the `voxel_grid`.
    - `filename`: The output filename.
    - `overwrite`: If `True`, overwrite an existing file.
    - `compression`: See `mrcfile.new` for documentation.
    """
    # Validate filename as MRC path and get suffix
    suffix = _validate_filename_and_return_suffix(filename)
    if suffix != ".mrc":
        raise OSError(
            f"The suffix for a volume in MRC format must be .mrc. Instead, got {suffix}."
        )
    if voxel_grid.ndim != 3:
        raise ValueError("voxel_grid.ndim must be equal to 3.")
    # Convert volume and voxel size to numpy arrays.
    voxel_grid_as_numpy = np.asarray(voxel_grid)
    voxel_size_as_numpy = np.asarray(voxel_size)
    # Create new file and write
    with mrcfile.new(filename, compression=compression, overwrite=overwrite) as mrc:
        mrc.set_data(voxel_grid_as_numpy)
        mrc.set_volume()
        mrc.voxel_size = voxel_size_as_numpy


def _validate_filename_and_return_suffix(filename: str | pathlib.Path):
    # Get suffixes
    suffixes = pathlib.Path(filename).suffixes
    # Make sure that leading suffix is valid MRC suffix
    if len(suffixes) == 0 or suffixes[0] not in [".mrc", ".mrcs"]:
        raise OSError(
            f"Filename should include .mrc or .mrcs suffix. Got filename {filename}."
        )
    return suffixes[0]



import warnings as _warnings
from typing import Any as _Any

from ._coordinates import (
    cartesian_to_polar as cartesian_to_polar,
    make_1d_coordinate_grid as make_1d_coordinate_grid,
    make_1d_frequency_grid as make_1d_frequency_grid,
    make_coordinate_grid as make_coordinate_grid,
    make_frequency_grid as make_frequency_grid,
    make_frequency_slice as make_frequency_slice,
    make_radial_coordinate_grid as make_radial_coordinate_grid,
    make_radial_frequency_grid as make_radial_frequency_grid,
)
from ._downsample import (
    block_reduce_downsample as block_reduce_downsample,
    fourier_crop_downsample as fourier_crop_downsample,
    fourier_crop_to_shape as fourier_crop_to_shape,
)
from ._edges import (
    crop_to_shape as crop_to_shape,
    pad_to_shape as pad_to_shape,
    resize_with_crop_or_pad as resize_with_crop_or_pad,
)
from ._fft import (
    fftn as fftn,
    ifftn as ifftn,
    irfftn as irfftn,
    rfftn as rfftn,
)
from ._fourier_statistics import (
    compute_binned_powerspectrum as compute_binned_powerspectrum,
    compute_fourier_ring_correlation as compute_fourier_ring_correlation,
    compute_fourier_shell_correlation as compute_fourier_shell_correlation,
)
from ._fourier_utils import (
    convert_fftn_to_rfftn as convert_fftn_to_rfftn,
    enforce_rfftn_self_conjugates as enforce_rfftn_self_conjugates,
)
from ._map_coordinates import (
    compute_spline_coefficients as compute_spline_coefficients,
    map_coordinates as map_coordinates,
    map_coordinates_spline as map_coordinates_spline,
)
from ._normalize import (
    background_subtract_image as background_subtract_image,
    compute_edge_value as compute_edge_value,
    rescale_fft as rescale_fft,
    rescale_image as rescale_image,
    standardize_fft as standardize_fft,
    standardize_image as standardize_image,
)
from ._operators import (
    AbstractFourierOperator as AbstractFourierOperator,
    AbstractRealOperator as AbstractRealOperator,
    CustomFourierOperator as CustomFourierOperator,
    FourierConstant as FourierConstant,
    FourierDC as FourierDC,
    FourierExp2D as FourierExp2D,
    FourierGaussian as FourierGaussian,
    FourierPhaseShifts as FourierPhaseShifts,
    FourierSinc as FourierSinc,
    PeakedFourierGaussian as PeakedFourierGaussian,
    RealConstant as RealConstant,
    RealGaussian as RealGaussian,
)
from ._radial_average import (
    compute_binned_radial_average as compute_binned_radial_average,
    radial_average_to_grid as radial_average_to_grid,
)
from ._rescale_pixel_size import (
    rescale_pixel_size as rescale_pixel_size,
)
from ._transforms import (
    AbstractFilter as AbstractFilter,
    AbstractImageTransform as AbstractImageTransform,
    AbstractMask as AbstractMask,
    CircularCosineMask as CircularCosineMask,
    CustomFilter as CustomFilter,
    CustomMask as CustomMask,
    Cylindrical2DCosineMask as Cylindrical2DCosineMask,
    HighpassFilter as HighpassFilter,
    LowpassFilter as LowpassFilter,
    PhaseShiftFFT as PhaseShiftFFT,
    Rectangular2DCosineMask as Rectangular2DCosineMask,
    Rectangular3DCosineMask as Rectangular3DCosineMask,
    RotateFFT as RotateFFT,
    ScaleImage as ScaleImage,
    SincCorrectionMask as SincCorrectionMask,
    SphericalCosineMask as SphericalCosineMask,
    SquareCosineMask as SquareCosineMask,
    WhiteningFilter as WhiteningFilter,
)


def __getattr__(name: str) -> _Any:
    # Future deprecations
    if name == "downsample_with_fourier_cropping":
        _warnings.warn(
            "'downsample_with_fourier_cropping' is deprecated"
            "has been renamed to 'fourier_crop_downsample'. "
            "The old name will be deprecated in cryoJAX 0.6.0.",
            category=FutureWarning,
            stacklevel=2,
        )
        return fourier_crop_downsample
    if name == "downsample_to_shape_with_fourier_cropping":
        _warnings.warn(
            "'downsample_to_shape_with_fourier_cropping' is deprecated"
            "has been renamed to 'fourier_crop_to_shape'. "
            "The old name will be deprecated in cryoJAX 0.6.0.",
            category=FutureWarning,
            stacklevel=2,
        )
        return fourier_crop_to_shape
    if name == "normalize_image":
        _warnings.warn(
            "'normalize_image' is deprecated and "
            "has been renamed to 'standardize_image'. "
            "The old name will be deprecated in cryoJAX 0.6.0.",
            category=FutureWarning,
            stacklevel=2,
        )
        return standardize_image
    if name == "operators":
        _warnings.warn(
            "Submodule `cryojax.ndimage.operators` is deprecated and "
            "has been moved to the `cryojax.ndimage` namespace. "
            "`cryojax.ndimage.operators` will be removed in cryoJAX 0.6.0.",
            category=FutureWarning,
            stacklevel=2,
        )
        from . import _operators as operators

        return operators
    if name == "transforms":
        _warnings.warn(
            "Submodule `cryojax.ndimage.transforms` is deprecated and "
            "has been moved to the `cryojax.ndimage` namespace. "
            "`cryojax.ndimage.transforms` will be removed in cryoJAX 0.6.0.",
            category=FutureWarning,
            stacklevel=2,
        )
        from . import _transforms as transforms

        return transforms

    raise AttributeError(f"cannot import name '{name}' from 'cryojax.ndimage'.")
